import json
import logging
from datetime import date, datetime
from gzip import GzipFile
from io import BytesIO

from dateutil.tz import tzutc
from requests import sessions

from hightouch.htevents.utils import remove_trailing_slash
from hightouch.htevents.version import VERSION

_session = sessions.Session()


def post(
    write_key,
    host=None,
    gzip=False,
    timeout=15,
    proxies=None,
    oauth_manager=None,
    **kwargs,
):
    """Post the `kwargs` to the API"""
    log = logging.getLogger('hightouch')
    body = kwargs
    if 'sentAt' not in body.keys():
        body['sentAt'] = datetime.utcnow().replace(tzinfo=tzutc()).isoformat()
    body['writeKey'] = write_key
    url = (
        remove_trailing_slash(host or 'https://us-east-1.hightouch-events.com')
        + '/v1/batch'
    )
    auth = None
    if oauth_manager:
        auth = oauth_manager.get_token()
    data = json.dumps(body, cls=DatetimeSerializer)
    log.debug('making request: %s', data)
    headers = {
        'Content-Type': 'application/json',
        'User-Agent': 'events-sdk-python/' + VERSION,
    }
    if auth:
        headers['Authorization'] = 'Bearer {}'.format(auth)

    if gzip:
        headers['Content-Encoding'] = 'gzip'
        buf = BytesIO()
        with GzipFile(fileobj=buf, mode='w') as gz:
            # 'data' was produced by json.dumps(),
            # whose default encoding is utf-8.
            gz.write(data.encode('utf-8'))
        data = buf.getvalue()

    kwargs = {
        'data': data,
        'headers': headers,
        'timeout': 15,
    }

    if proxies:
        kwargs['proxies'] = proxies
    res = None
    try:
        res = _session.post(url, data=data, headers=headers, timeout=timeout)
    except Exception as e:
        log.error(e)
        raise e

    if res.status_code == 200:
        log.debug('data uploaded successfully')
        return res

    if res.status_code == 401:
        log.error('Unknown error: [%s] %s', res.status_code, res.reason)
        raise APIError(res.status_code, 'unknown', res.text)

    if oauth_manager and res.status_code in [400, 401, 403]:
        oauth_manager.clear_token()

    try:
        payload = res.json()
        log.debug('received response: %s', payload)
        raise APIError(res.status_code, payload['code'], payload['message'])
    except ValueError:
        log.error('Unknown error: [%s] %s', res.status_code, res.reason)
        raise APIError(res.status_code, 'unknown', res.text)


class APIError(Exception):
    def __init__(self, status, code, message):
        self.message = message
        self.status = status
        self.code = code

    def __str__(self):
        msg = '[Hightouch] {0}: {1} ({2})'
        return msg.format(self.code, self.message, self.status)


class DatetimeSerializer(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()

        return json.JSONEncoder.default(self, obj)
